import logging

import cc1101


logging.basicConfig(level=logging.INFO)

with cc1101.CC1101() as transceiver:
    transceiver.set_base_frequency_hertz(433.5e6)
    transceiver.set_symbol_rate_baud(600)
    transceiver.set_sync_mode(cc1101.SyncMode.NO_PREAMBLE_AND_SYNC_WORD)
    transceiver.set_packet_length_mode(cc1101.PacketLengthMode.FIXED)
    transceiver.set_packet_length_bytes(4)
    transceiver.disable_checksum()
    print(transceiver)
    transceiver.transmit(b"\xff\x00\xaa\xff")



# python-cc1101 - Python Library to Transmit RF Signals via C1101 Transceivers
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import logging
import sys

import cc1101
import cc1101.options

_LOGGER = logging.getLogger(__name__)


def _transmit():
    argparser = argparse.ArgumentParser(
        description="Transmits the payload provided via standard input (stdin)"
        " OOK-modulated in big-endian bit order.",
        allow_abbrev=False,
    )
    argparser.add_argument("-f", "--base-frequency-hertz", type=int)
    argparser.add_argument("-r", "--symbol-rate-baud", type=int)
    argparser.add_argument(
        "-s",
        "--sync-mode",
        type=str,
        choices=[m.name.lower().replace("_", "-") for m in cc1101.options.SyncMode],
    )
    argparser.add_argument(
        "-l",
        "--packet-length-mode",
        type=str,
        choices=[m.name.lower() for m in cc1101.options.PacketLengthMode],
    )
    argparser.add_argument("--disable-checksum", action="store_true")
    argparser.add_argument("-d", "--debug", action="store_true")
    args = argparser.parse_args()
    logging.basicConfig(
        level=logging.DEBUG if args.debug else logging.INFO,
        # format="%(asctime)s:%(levelname)s:%(name)s:%(funcName)s:%(message)s",
        format="%(name)s:%(funcName)s:%(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S%z",
    )
    _LOGGER.debug("args=%r", args)
    with cc1101.CC1101() as transceiver:
        if args.base_frequency_hertz:
            transceiver.set_base_frequency_hertz(args.base_frequency_hertz)
        if args.symbol_rate_baud:
            transceiver.set_symbol_rate_baud(args.symbol_rate_baud)
        if args.sync_mode:
            transceiver.set_sync_mode(
                cc1101.options.SyncMode[args.sync_mode.upper().replace("-", "_")]
            )
        payload = sys.stdin.buffer.read()
        if args.packet_length_mode:
            packet_length_mode = cc1101.options.PacketLengthMode[
                args.packet_length_mode.upper()
            ]
            # default: variable length
            transceiver.set_packet_length_mode(packet_length_mode)
            # default: 255 (maximum)
            if packet_length_mode == cc1101.options.PacketLengthMode.FIXED:
                transceiver.set_packet_length_bytes(len(payload))
        if args.disable_checksum:
            transceiver.disable_checksum()
        _LOGGER.debug("transceiver=%r", transceiver)
        transceiver.transmit(payload)



import enum


class StrobeAddress(enum.IntEnum):
    # see "Table 42: Command Strobes"
    SRES = 0x30
    SFSTXON = 0x31
    SXOFF = 0x32
    SCAL = 0x33
    SRX = 0x34
    STX = 0x35
    SIDLE = 0x36
    SWOR = 0x38
    SPWD = 0x39
    SFRX = 0x3A
    SFTX = 0x3B
    SWORRST = 0x3C
    SNOP = 0x3D


class ConfigurationRegisterAddress(enum.IntEnum):
    # see "Table 43: Configuration Registers Overview"
    IOCFG2 = 0x00  # GDO2 output pin configuration
    IOCFG1 = 0x01  # GDO1 output pin configuration
    IOCFG0 = 0x02  # GDO0 output pin configuration
    FIFOTHR = 0x03  # RX FIFO and TX FIFO thresholds
    SYNC1 = 0x04  # Sync word, high byte
    SYNC0 = 0x05  # Sync word, low byte
    PKTLEN = 0x06  # Packet length
    PKTCTRL1 = 0x07  # Packet automation control
    PKTCTRL0 = 0x08  # Packet automation control
    ADDR = 0x09  # Device address
    CHANNR = 0x0A  # Channel number
    FSCTRL1 = 0x0B  # Frequency synthesizer control
    FSCTRL0 = 0x0C  # Frequency synthesizer control
    FREQ2 = 0x0D  # Frequency control word, high byte
    FREQ1 = 0x0E  # Frequency control word, middle byte
    FREQ0 = 0x0F  # Frequency control word, low byte
    MDMCFG4 = 0x10  # Modem configuration
    MDMCFG3 = 0x11  # Modem configuration
    MDMCFG2 = 0x12  # Modem configuration
    MDMCFG1 = 0x13  # Modem configuration
    MDMCFG0 = 0x14  # Modem configuration
    DEVIATN = 0x15  # Modem deviation setting
    MCSM2 = 0x16  # Main Radio Control State Machine configuration
    MCSM1 = 0x17  # Main Radio Control State Machine configuration
    MCSM0 = 0x18  # Main Radio Control State Machine configuration
    FOCCFG = 0x19  # Frequency Offset Compensation configuration
    BSCFG = 0x1A  # Bit Synchronization configuration
    AGCTRL2 = 0x1B  # AGC control
    AGCTRL1 = 0x1C  # AGC control
    AGCTRL0 = 0x1D  # AGC control
    WOREVT1 = 0x1E  # High byte Event 0 timeout
    WOREVT0 = 0x1F  # Low byte Event 0 timeout
    WORCTRL = 0x20  # Wake On Radio control
    FREND1 = 0x21  # Front end RX configuration
    FREND0 = 0x22  # Front end TX configuration
    FSCAL3 = 0x23  # Frequency synthesizer calibration
    FSCAL2 = 0x24  # Frequency synthesizer calibration
    FSCAL1 = 0x25  # Frequency synthesizer calibration
    FSCAL0 = 0x26  # Frequency synthesizer calibration
    RCCTRL1 = 0x27  # RC oscillator configuration
    RCCTRL0 = 0x28  # RC oscillator configuration
    FSTEST = 0x29  # Frequency synthesizer calibration control
    PTEST = 0x2A  # Production test
    AGCTEST = 0x2B  # AGC test
    TEST2 = 0x2C  # Various test settings
    TEST1 = 0x2D  # Various test settings
    TEST0 = 0x2E  # Various test settings


class StatusRegisterAddress(enum.IntEnum):
    # see "Table 44: Status Registers Overview"
    PARTNUM = 0x30  # Part number for CC1101
    VERSION = 0x31  # Current version number
    FREQEST = 0x32  # Frequency Offset Estimate
    LQI = 0x33  # Demodulator estimate for Link Quality
    RSSI = 0x34  # Received signal strength indication
    MARCSTATE = 0x35  # Control state machine state
    WORTIME1 = 0x36  # High byte of WOR timer
    WORTIME0 = 0x37  # Low byte of WOR timer
    PKTSTATUS = 0x38  # Current GDOx status and packet status
    VCO_VC_DAC = 0x39  # Current setting from PLL calibration module
    TXBYTES = 0x3A  # Underflow and number of bytes in the TX FIFO
    RXBYTES = 0x3B  # Overflow and number of bytes in the RX FIFO
    RCCTRL1_STATUS = 0x3C  # Last RC oscillator calibration result
    RCCTRL0_STATUS = 0x3D  # Last RC oscillator calibration result


class FIFORegisterAddress(enum.IntEnum):
    # see "10.5 FIFO Access"
    # > When the R/W-bit is zero, the TX FIFO is accessed,
    # > and the RX FIFO is accessed when the R/W-bit is one.
    TX = RX = 0x3F



# python-cc1101 - Python Library to Transmit RF Signals via C1101 Transceivers
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import io
import logging
import unittest.mock

import pytest

import cc1101._cli
from cc1101.options import PacketLengthMode, SyncMode

# pylint: disable=protected-access


@pytest.mark.parametrize(
    (
        "args",
        "base_frequency_hertz",
        "symbol_rate_baud",
        "sync_mode",
        "packet_length_mode",
        "checksum_disabled",
    ),
    (
        ([""], None, None, None, None, False),
        (["", "-f", "433920000"], 433920000, None, None, None, False),
        (
            ["", "--base-frequency-hertz", "868000000"],
            868000000,
            None,
            None,
            None,
            False,
        ),
        (["", "-r", "10000"], None, 10000, None, None, False),
        (["", "--symbol-rate-baud", "12345"], None, 12345, None, None, False),
        (
            ["", "-s", "no-preamble-and-sync-word"],
            None,
            None,
            SyncMode.NO_PREAMBLE_AND_SYNC_WORD,
            None,
            False,
        ),
        (
            ["", "--sync-mode", "no-preamble-and-sync-word"],
            None,
            None,
            SyncMode.NO_PREAMBLE_AND_SYNC_WORD,
            None,
            False,
        ),
        (
            ["", "--sync-mode", "transmit-16-match-15-bits"],
            None,
            None,
            SyncMode.TRANSMIT_16_MATCH_15_BITS,
            None,
            False,
        ),
        (["", "-l", "fixed"], None, None, None, PacketLengthMode.FIXED, False),
        (
            ["", "--packet-length-mode", "fixed"],
            None,
            None,
            None,
            PacketLengthMode.FIXED,
            False,
        ),
        (
            ["", "--packet-length-mode", "variable"],
            None,
            None,
            None,
            PacketLengthMode.VARIABLE,
            False,
        ),
        (["", "--disable-checksum"], None, None, None, None, True),
    ),
)
@pytest.mark.parametrize("payload", (b"", b"message"))
def test_configure_device(
    args,
    base_frequency_hertz,
    symbol_rate_baud,
    sync_mode,
    packet_length_mode,
    checksum_disabled,
    payload,
):
    # pylint: disable=too-many-arguments
    with unittest.mock.patch("cc1101.CC1101") as transceiver_mock:
        stdin_mock = unittest.mock.MagicMock()
        stdin_mock.buffer = io.BytesIO(payload)
        with unittest.mock.patch("sys.stdin", stdin_mock):
            with unittest.mock.patch("sys.argv", args):
                cc1101._cli._transmit()
    transceiver_mock.assert_called_once_with()
    if base_frequency_hertz is None:
        transceiver_mock().__enter__().set_base_frequency_hertz.assert_not_called()
    else:
        transceiver_mock().__enter__().set_base_frequency_hertz.assert_called_once_with(
            base_frequency_hertz
        )
    if symbol_rate_baud is None:
        transceiver_mock().__enter__().set_symbol_rate_baud.assert_not_called()
    else:
        transceiver_mock().__enter__().set_symbol_rate_baud.assert_called_once_with(
            symbol_rate_baud
        )
    if sync_mode is None:
        transceiver_mock().__enter__().set_sync_mode.assert_not_called()
    else:
        transceiver_mock().__enter__().set_sync_mode.assert_called_once_with(sync_mode)
    if packet_length_mode is None:
        transceiver_mock().__enter__().set_packet_length_mode.assert_not_called()
    else:
        transceiver_mock().__enter__().set_packet_length_mode.assert_called_once_with(
            packet_length_mode
        )
    if packet_length_mode == PacketLengthMode.FIXED:
        transceiver_mock().__enter__().set_packet_length_bytes.assert_called_with(
            len(payload)
        )
    else:
        transceiver_mock().__enter__().set_packet_length_bytes.assert_not_called()
    if checksum_disabled:
        transceiver_mock().__enter__().disable_checksum.assert_called_once_with()
    else:
        transceiver_mock().__enter__().disable_checksum.assert_not_called()


@pytest.mark.parametrize(
    ("args", "root_log_level"), (([""], logging.INFO), (["", "--debug"], logging.DEBUG))
)
def test_root_log_level(args, root_log_level):
    stdin_mock = unittest.mock.MagicMock()
    stdin_mock.buffer = io.BytesIO(b"")
    with unittest.mock.patch("cc1101.CC1101"), unittest.mock.patch(
        "sys.stdin", stdin_mock
    ), unittest.mock.patch("sys.argv", args), unittest.mock.patch(
        "logging.basicConfig"
    ) as logging_basic_config_mock:
        cc1101._cli._transmit()
    assert logging_basic_config_mock.call_count == 1
    assert logging_basic_config_mock.call_args[1]["level"] == root_log_level



import unittest.mock

import pytest

import cc1101.options

# pylint: disable=protected-access


def test_transmit_empty_payload(transceiver):
    with unittest.mock.patch.object(
        transceiver,
        "get_packet_length_mode",
        return_value=cc1101.options.PacketLengthMode.VARIABLE,
    ), unittest.mock.patch.object(
        transceiver, "get_packet_length_bytes", return_value=21
    ):
        with pytest.raises(ValueError, match=r"\bempty\b"):
            transceiver.transmit([])


@pytest.mark.parametrize(
    ("max_packet_length", "payload"),
    ((3, "\x04\x01\x02\x03"), (4, [7, 21, 42, 0, 0, 1, 2, 3])),
)
def test_transmit_exceeding_max_length(transceiver, max_packet_length, payload):
    with unittest.mock.patch.object(
        transceiver,
        "get_packet_length_mode",
        return_value=cc1101.options.PacketLengthMode.VARIABLE,
    ), unittest.mock.patch.object(
        transceiver, "get_packet_length_bytes", return_value=max_packet_length
    ):
        with pytest.raises(
            ValueError, match=r"\bpayload exceeds maximum payload length\b"
        ):
            transceiver.transmit(payload)


@pytest.mark.parametrize(
    ("packet_length", "payload"),
    ((3, "\x04\x01\x02\x03"), (4, [7, 21, 42, 0, 0, 1, 2, 3])),
)
def test_transmit_unexpected_payload_len(transceiver, packet_length, payload):
    with unittest.mock.patch.object(
        transceiver,
        "get_packet_length_mode",
        return_value=cc1101.options.PacketLengthMode.FIXED,
    ), unittest.mock.patch.object(
        transceiver, "get_packet_length_bytes", return_value=packet_length
    ):
        with pytest.raises(ValueError, match=r"\bpayload length\b"):
            transceiver.transmit(payload)


@pytest.mark.parametrize("payload", (b"\0", b"\xaa\xbb\xcc", bytes(range(42))))
def test_transmit_fixed(transceiver, payload):
    transceiver._spi.xfer.side_effect = lambda v: [15] * len(v)
    with unittest.mock.patch.object(
        transceiver,
        "get_packet_length_mode",
        return_value=cc1101.options.PacketLengthMode.FIXED,
    ), unittest.mock.patch.object(
        transceiver, "get_packet_length_bytes", return_value=len(payload)
    ), unittest.mock.patch.object(
        transceiver,
        "get_main_radio_control_state_machine_state",
        return_value=cc1101.MainRadioControlStateMachineState.IDLE,
    ):
        transceiver.transmit(payload)
    assert transceiver._spi.xfer.call_args_list == [
        unittest.mock.call([0x3B]),  # flush
        unittest.mock.call([0x3F | 0x40] + list(payload)),
        unittest.mock.call([0x35]),  # start transmission
    ]


@pytest.mark.parametrize(
    "payload", (b"\x01\0", b"\x03\xaa\xbb\xcc", b"\x10" + bytes(range(16)))
)
def test_transmit_variable(transceiver, payload):
    transceiver._spi.xfer.side_effect = lambda v: [15] * len(v)
    with unittest.mock.patch.object(
        transceiver,
        "get_packet_length_mode",
        return_value=cc1101.options.PacketLengthMode.VARIABLE,
    ), unittest.mock.patch.object(
        transceiver, "get_packet_length_bytes", return_value=255
    ), unittest.mock.patch.object(
        transceiver,
        "get_main_radio_control_state_machine_state",
        return_value=cc1101.MainRadioControlStateMachineState.IDLE,
    ):
        transceiver.transmit(payload)
    assert transceiver._spi.xfer.call_args_list == [
        unittest.mock.call([0x3B]),  # flush
        unittest.mock.call([0x3F | 0x40] + [len(payload)] + list(payload)),
        unittest.mock.call([0x35]),  # start transmission
    ]



import unittest.mock

import pytest

# pylint: disable=protected-access


def test__enable_receive_mode(transceiver):
    transceiver._spi.xfer.return_value = [15]
    transceiver._enable_receive_mode()
    transceiver._spi.xfer.assert_called_once_with([0x34 | 0x00])


@pytest.mark.parametrize("data", [b"\0", b"\x12\x45\x56"])
def test__get_received_packet(transceiver, data):
    fifo_buffer = list(data) + [128, (1 << 7) | 42]
    with unittest.mock.patch.object(
        transceiver, "_read_status_register", return_value=len(fifo_buffer)
    ) as read_status_register, unittest.mock.patch.object(
        transceiver, "_read_burst", return_value=fifo_buffer
    ) as read_burst_mock:
        received_packet = transceiver._get_received_packet()
    read_status_register.assert_called_once_with(0x3B)
    read_burst_mock.assert_called_once_with(
        start_register=0x3F, length=len(fifo_buffer)
    )
    assert received_packet.data == data
    assert received_packet._rssi_index == 128
    assert received_packet.checksum_valid
    assert received_packet.link_quality_indicator == 42



# python-cc1101 - Python Library to Transmit RF Signals via C1101 Transceivers
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import contextlib
import unittest.mock

import pytest

import cc1101
import cc1101.addresses

# pylint: disable=protected-access


@pytest.fixture(scope="function")
def spidev_mock(tmp_path):
    class _SpiDevMock:
        def __init__(self):
            self._file = None

        def open(self, bus, select):
            path = tmp_path.joinpath("spidev{}.{}~".format(bus, select))
            self._file = path.open("w+")

        def fileno(self):
            # mimic behaviour of spidev.SpiDev.fileno()
            return self._file.fileno() if self._file else -1

        def close(self):
            self._file.close()
            self._file = None  # for fileno

    return _SpiDevMock


# pylint: disable=redefined-outer-name; using fixture


@contextlib.contextmanager
def _mock_hardware_access():
    with unittest.mock.patch.object(
        cc1101.CC1101, "_reset"
    ), unittest.mock.patch.object(
        cc1101.CC1101, "_verify_chip"
    ), unittest.mock.patch.object(
        cc1101.CC1101, "_configure_defaults"
    ), unittest.mock.patch.object(
        cc1101.CC1101,
        "get_main_radio_control_state_machine_state",
        return_value=cc1101.MainRadioControlStateMachineState.IDLE,
    ):
        yield


def test_context_lock(spidev_mock):
    with _mock_hardware_access():
        with unittest.mock.patch("spidev.SpiDev", spidev_mock):
            transceiver = cc1101.CC1101(lock_spi_device=True)
        with transceiver:
            with unittest.mock.patch("spidev.SpiDev", spidev_mock):
                transceiver2 = cc1101.CC1101(lock_spi_device=True)
            with pytest.raises(BlockingIOError):
                with transceiver2:
                    pass
            with unittest.mock.patch("spidev.SpiDev", spidev_mock):
                transceiver3 = cc1101.CC1101(lock_spi_device=False)
            with transceiver3:
                pass
        with transceiver2:  # closing unlocks
            pass


def test_context_no_lock(spidev_mock):
    with _mock_hardware_access():
        with unittest.mock.patch("spidev.SpiDev", spidev_mock):
            transceiver = cc1101.CC1101(lock_spi_device=False)
        with transceiver:
            with unittest.mock.patch("spidev.SpiDev", spidev_mock):
                transceiver2 = cc1101.CC1101(lock_spi_device=True)
            with transceiver2:
                pass


def test_unlock_spi_device(spidev_mock):
    with _mock_hardware_access():
        with unittest.mock.patch("spidev.SpiDev", spidev_mock):
            transceiver = cc1101.CC1101(lock_spi_device=True)
            transceiver2 = cc1101.CC1101(lock_spi_device=True)
        with transceiver:  # acquire lock
            with pytest.raises(BlockingIOError):
                with transceiver2:
                    pass
            transceiver.unlock_spi_device()
            with transceiver2:
                pass


def test_unlock_spi_device_double(spidev_mock):
    with _mock_hardware_access():
        with unittest.mock.patch("spidev.SpiDev", spidev_mock):
            transceiver = cc1101.CC1101(lock_spi_device=True)
        # verify no error occurs
        with transceiver:  # acquire lock
            transceiver.unlock_spi_device()
            transceiver.unlock_spi_device()


def test_unlock_spi_device_outside_context(spidev_mock):
    with _mock_hardware_access():
        with unittest.mock.patch("spidev.SpiDev", spidev_mock):
            transceiver = cc1101.CC1101(lock_spi_device=True)
        # verify no error occurs
        transceiver.unlock_spi_device()
        with transceiver:  # acquire lock
            pass
        transceiver.unlock_spi_device()


def test_unlock_spi_device_no_lock(spidev_mock):
    with _mock_hardware_access():
        with unittest.mock.patch("spidev.SpiDev", spidev_mock):
            transceiver = cc1101.CC1101(lock_spi_device=False)
        with transceiver:  # no lock acquired
            # verify no error occurs
            transceiver.unlock_spi_device()
